#include <ci/FFT.H>
#include <ci/registration.H>
#include "ui.h"





ci::Area::Filter_Function filter_functions[12] = {
  ci::no_filter,                  // 0
  ci::high_pass_linear_filter,    // 1
  ci::low_pass_linear_filter,     // 2
  ci::high_pass_quadratic_filter, // 3
  ci::low_pass_quadratic_filter,  // 4
  ci::high_pass_filter_1,         // 5
  ci::low_pass_filter_1,          // 6
  ci::high_pass_filter_2,         // 7
  ci::low_pass_filter_2,          // 8
  ci::high_pass_filter_4,         // 9
  ci::low_pass_filter_4           // 10
};

int registration_in_halfspace = 1;

int registration_correlation_type = 1;

float min_registration_emphasis_radius = 0.35;
float registration_emphasis_radius_factor = 1.2;

int registration_rgb_filter_1 = 1;
int registration_rgb_filter_2 = 2;
float registration_rgb_filter_radius_1 = 1;
float registration_rgb_filter_radius_2 = 0.25;
int registration_topography_filter_1 = 1;
int registration_topography_filter_2 = 6;
float registration_topography_filter_radius_1 = 1.2;
float registration_topography_filter_radius_2 = 0.3;
int registration_iso_emphasis = 1;

// order: (blue = 1), green, red, topography
float registration_weighting_coefficients[] = {0.2, 0.12, 0.01};

/*

int registration_in_halfspace = 0;

int registration_correlation_type = 1;

float min_registration_emphasis_radius = 0.35;  // wa2 35
float registration_emphasis_radius_factor = 1.2;

int registration_rgb_filter_1 = 1;
int registration_rgb_filter_2 = 2;
float registration_rgb_filter_radius_1 = 0.3; // was 1;
float registration_rgb_filter_radius_2 = 0.005;// was0.25;
int registration_topography_filter_1 = 2; // was 1;
int registration_topography_filter_2 = 6;
float registration_topography_filter_radius_1 = 0.25; // was 1.2;
float registration_topography_filter_radius_2 = 0.02; // was 0.05;
int registration_iso_emphasis = 1;

// order: (blue = 1), green, red, topography
float registration_weighting_coefficients[] = {0.2, 0.12, 0.02}; // was {0.2, 0.12, 0.01}
*/


// gauss table of radius n/3;
const float * gauss_table(int n) {
  static float * table = 0;
  static int n_ = 0;
  if (n != n_) {
    float par = -9.0 / float(n * n);
    n_ = n;
    delete[] table;
    table = new float[n];
    float * t = table;
    for (int i = 0; i < n; i++)
      *t++ = exp(par * (i * i));
  }
  return table;
}

const int table_size = 1024;

void init_gauss_filter();

void gauss_filter(float * data, int x, int y, float cx, float cy, float rx, float ry, int dx = 0) {
  if (!dx)
    dx = x;

  const float * t = gauss_table(table_size);
  float pary = float(table_size) / (3.0 * ry);
  float parx = float(table_size) / (3.0 * rx);
  for (int j = 0; j < y; j++) {
    int index = int(pary * fabs(cy - j) + 0.5);
    float fy = (index < table_size) ? t[index] : 0.0;
    float * d = data + j * dx;
    for (int i = 0; i < x; i++) {
      index = int(parx * fabs(cx - i) + 0.5);
      float fx = (index < table_size) ? t[index] : 0.0;
      *d = *d * fx * fy;
      d++;

    }
  }

}




int Live_Window::register_image(ci::Selection * sel, float * sel_window_array, float * &array, int &array_size,
                                ci::FFT2 &fft2, ci::FFT2 &fft2p, ci::Area &m1, ci::Area &m2, ci::Area &m3, ci::Area &m4, ci::Area &m5, float xx, float yy, float ww, float hh) {
  Lock lock(image_data);
  ci::Image * im = image_data.fast_value();
  if (!im) return 0;
  int x = im->x();
  int y = im->y();
  if (x*y*3 > array_size) {
    array_size = x * y * 4;
    delete[] array;
    array = new float[array_size];
  }
  float *r = array;
  float *g = array + x * y;
  float *b = array + 2 * x * y;
  float *p = array + 3 * x * y;
  if (!im->update_rgb(r, g, b, x))
    return 0;

  im->update_surface_fft(&m1 , &fft2);
  m1.set(im->surface_fft());
  //m1.filter(&ci::low_pass_filter_1, 0.3);
  //m1.filter(&ci::high_pass_filter_1, 32.0/float(m1.x()));
  fft2(&m1, -1);
  m1.copy_into(p, x, y);

  ci::hann_filter(r, x, y, x);
  ci::hann_filter(g, x, y, x);
  ci::hann_filter(b, x, y, x);
  ci::hann_filter(p, x, y, x);

  gauss_filter(r, x, y, xx, yy, ww, hh);
  gauss_filter(g, x, y, xx, yy, ww, hh);
  gauss_filter(b, x, y, xx, yy, ww, hh);
  gauss_filter(p, x, y, xx, yy, ww, hh);




  //ci::circular_hann_filter(r, x, y, x);
  //ci::circular_hann_filter(g, x, y, x);
  //ci::circular_hann_filter(b, x, y, x);
  //ci::circular_hann_filter(p, x, y, x);

  float X, Y, W, H;
  sel->size(X, Y, W, H);
  float cx = X + W / 2;
  float cy = Y + H / 2;
  const float * sel_images[4];
  sel_images[0] = sel_window_array;
  sel_images[1] = sel_window_array + x * y;
  sel_images[2] = sel_window_array + 2 * x * y;
  sel_images[3] = sel_window_array + 3 * x * y;
  const float * images[4];

  images[0] = b;
  images[1] = g;
  images[2] = r;
  images[3] = p;

  ci::Area::Filter_Function registration_filters_1[4];
  registration_filters_1[0] = registration_filters_1[1] = registration_filters_1[2] = filter_functions[registration_rgb_filter_1];
  registration_filters_1[3] = filter_functions[registration_topography_filter_1];
  ci::Area::Filter_Function registration_filters_2[4];
  registration_filters_2[0] = registration_filters_2[1] = registration_filters_2[2] = filter_functions[registration_rgb_filter_2];
  registration_filters_2[3] = filter_functions[registration_topography_filter_2];

  float registration_filters_values_1[4];
  registration_filters_values_1[0] = registration_filters_values_1[1] = registration_filters_values_1[2] = registration_rgb_filter_radius_1;
  registration_filters_values_1[3] = registration_topography_filter_radius_1;
  float registration_filters_values_2[4];
  registration_filters_values_2[0] = registration_filters_values_2[1] = registration_filters_values_2[2] = registration_rgb_filter_radius_2;
  registration_filters_values_2[3] = registration_topography_filter_radius_2;



  ci::Matrix mm = ci::get_registration_matrix(4, x, y, x, 0, images, sel_images, registration_weighting_coefficients, x * cx, y * cy, m1, m2, m3, m4, m5, fft2, fft2p, registration_in_halfspace,
                  registration_filters_1, registration_filters_2, registration_filters_values_1, registration_filters_values_2, registration_correlation_type);
  ci::Matching_Vector  m_vector;
  m_vector.valid = 2;
  float XX = 0;
  float YY = 0;
  mm(XX, YY);
  m_vector.x1 = XX / float(x);
  m_vector.y1 = YY / float(y);
  XX = 1;
  YY = 1;
  mm(XX, YY);
  m_vector.x2 = XX / float(x);
  m_vector.y2 = YY / float(y);
  match_vector.set_new(m_vector);
  Flat_Window * gw = gl_window;
  if (!gw->view_mode())
    gw->redraw_general_overlay();
  //match_vector = m_vector;
  return 1;
}
//static Fl_Color reg_sel_color = FL_GRAY;

int Live_Window::remove_auto_registration() {
  if(match_vector.value().valid==2){
    //registration_ctl->selection_color(fl_lighter(FL_GRAY));
    registration_ctl->value(0);
    registration_ctl->selection_color(fl_lighter(FL_GRAY));
    registration_ctl->redraw();
    ci::Matching_Vector  m_vector;
    m_vector.valid = 0;
    match_vector.set_new(m_vector);
    Flat_Window * gw = gl_window;
    if (!gw->view_mode())
      gw->redraw_general_overlay();
    return 1;
  }
  return 0;
}

int Live_Window::remove_vector_matching(bool update) {
  if(match_vector.value().valid==1){
    //registration_ctl->selection_color(fl_lighter(FL_GRAY));
    vector_ctl->value(0);
    vector_ctl->selection_color(fl_lighter(FL_GRAY));
    vector_ctl->redraw();
    if(update){
      ci::Matching_Vector  m_vector;
      m_vector.valid = 0;
      match_vector.set_new(m_vector);
      Flat_Window * gw = gl_window;
      if (!gw->view_mode())
        gw->redraw_general_overlay();
    }
    return 1;
  }
  return 0;
}


int Live_Window::register_image() {
  if(remove_auto_registration())
    return 0;
   if(gl_window->in_selection_mode == ci::Selection::POLYGON){
      gl_window->clear_selection_state(ci::Selection::POLYGON);
      if(freeform_ctl->value()){
        freeform_ctl->value(0);
        freeform_ctl->redraw();
      }
   }else if(gl_window->in_selection_mode == ci::Selection::CIRCLE){
      gl_window->clear_selection_state(ci::Selection::CIRCLE);
      if(circle_ctl->value()){
        circle_ctl->value(0);
        circle_ctl->redraw();
      }
   }else if(gl_window->in_selection_mode == ci::Selection::LINE){
      gl_window->clear_selection_state(ci::Selection::LINE);
      if(strip_ctl->value()){
        strip_ctl->value(0);
        strip_ctl->redraw();
      }
   }

   Fl::flush();


  //gl_window->set_selection_state(0);

  ci::Selection_Task selection_task(ci::current_selection_task());
  if (!selection_task.selection) return 0;
  if (!selection_task.selection->valid()) return 0;
  remove_vector_matching(0);

  //registration_ctl->value(1);

  float xx, yy, ww, hh;
  selection_task.selection->size(xx, yy, ww, hh);
  xx += ww / 2;
  yy += hh / 2;

  ww *= 0.5;
  hh *= 0.5;

  if (registration_iso_emphasis){
    if(ww < hh)
      ww = hh;
    else
      hh = ww;

  }
  if (ww < min_registration_emphasis_radius)
    ww = min_registration_emphasis_radius;
  if (hh < min_registration_emphasis_radius)
    hh = min_registration_emphasis_radius;


  Live_Window * win = selection_task.window;
  if (!win || win == this)
    return 0;
  Lock lock(win->image_data);
  ci::Image * sel_im = win->image_data.fast_value();


  if (!sel_im)
    return 0;
  if(vector_ctl->value()){
    vector_ctl->value(0);
    vector_ctl->redraw();
  }
  registration_ctl->value(1);
  registration_ctl->selection_color(fl_lighter(FL_GRAY));
  registration_ctl->redraw();
  cursor(FL_CURSOR_WAIT);
  Fl::flush();
  int x = sel_im->x();
  int y = sel_im->y();
  xx *= x;
  yy *= y;
  ww *= x;
  hh *= y;
  ci::Image * im = image_data.fast_value();
  if (im->x() != x || im->y() != y)
    return 0;
  int sel_array_size = x * y * 4;
  float * sel_array = new float[sel_array_size];
  float *b = sel_array;
  float *g = sel_array + x * y;
  float *r = sel_array + 2 * x * y;
  float *p = sel_array + 3 * x * y;
  int ret = 0;

  if (sel_im->update_rgb(r, g, b, x)) {
    ci::FFT2 fft2;
    ci::FFT2 fft2p;
    ci::Area m1;
    ci::Area m2;
    ci::Area m3;
    ci::Area m4;
    ci::Area m5;
    sel_im->update_surface_fft(&m1 , &fft2);
    m1.set(sel_im->surface_fft());
    m1.filter(&ci::low_pass_filter_2, 0.2);
    //m1.filter(&ci::high_pass_filter_1, 32.0/float(m1.x()));

    fft2(&m1, -1);
    m1.copy_into(p, x, y);

    //ci::hann_filter(r, x, y, x);
    ci::hann_filter(g, x, y, x);
    ci::hann_filter(b, x, y, x);
    ci::hann_filter(p, x, y, x);

    gauss_filter(r, x, y, xx, yy, ww, hh);
    gauss_filter(g, x, y, xx, yy, ww, hh);
    gauss_filter(b, x, y, xx, yy, ww, hh);
    gauss_filter(p, x, y, xx, yy, ww, hh);

    //ci::circular_hann_filter(r, x, y, x);
    //ci::circular_hann_filter(g, x, y, x);
    //ci::circular_hann_filter(b, x, y, x);
    //ci::circular_hann_filter(p, x, y, x);

    float * array = 0;
    int array_size = 0;
    ret = register_image(selection_task.selection, sel_array, array, array_size, fft2, fft2p, m1, m2, m3, m4, m5, xx, yy, ww * 1.2, hh * 1.2);
    delete[] array;
  }
  delete[] sel_array;

  //reg_sel_color = registration_ctl->selection_color();
  registration_ctl->selection_color(FL_WHITE);
  registration_ctl->redraw();
  cursor(FL_CURSOR_DEFAULT);
  return ret;
}






